#include "Ogre.h"
#include "Noise.h"
#include "erodefault.h"
#include "watererode.h"

#include "TerrainMesh.h"

class CFTerr
{
    private:

	    CFractal frac;
        ErodeFault erfau;
        WaterErode watero;

	    float yscale, fzoom, foctaves, foffset, fgain;
	    float xoff, zoff;
	    int ibasis;

        int mExtent;
        TerrainMesh *tMesh;
        float *mheightData;

		Entity *mEnt;
		SceneNode *mSn;

    public:
        //----------------------------------------------------------------------
        CFTerr()
	    {
		    frac.Init(2, 414, 0.5f, 2.4f);
		    yscale = 200.0f;
		    fzoom = 0.001f;
		    foctaves = 8.0f;
		    foffset = 0.57f;
		    fgain = 3.5f;
		    ibasis = 5;
		    xoff = 3.02f;
		    zoff = -8.08f;
		    // Ridged Offset=0.57 Gain=3.5
            
            tMesh = 0;
            mheightData = 0;
	    }
        //----------------------------------------------------------------------
        ~CFTerr()
	    {
            delete tMesh;
            delete [] mheightData;
	    }
        //----------------------------------------------------------------------
        void SetBasisFunc(const int ival)	{	ibasis = ival;	}
        //----------------------------------------------------------------------
        void SetYscale(const float fval)	{	yscale = fval;	}
        //----------------------------------------------------------------------
        void SetZoom(const float fval)	{	fzoom = (float)fval*(0.001f/1024.0f);	}
        //----------------------------------------------------------------------
        void SetOctaves(const float fval)	{	foctaves = fval;	}
        //----------------------------------------------------------------------
        void SetOffset(const float fval)	{	foffset = fval;	}
        //----------------------------------------------------------------------
        void SetGain(const float fval)	{	fgain = fval;	}
        //----------------------------------------------------------------------
        void AddXoff(const float fval)	{	xoff += fzoom*fval;	}
        //----------------------------------------------------------------------
        void AddZoff(const float fval)	{	zoff += fzoom*fval;	}
        //----------------------------------------------------------------------
        void SetSeed(const unsigned int iseed)
	    {
		    frac.Init(2, iseed, 0.5f, 2.4f);
	    }
        //----------------------------------------------------------------------
        void NewTerrain(const String &entname, SceneManager *SceneMgr, int extents)
	    {

            mExtent = extents;
            
    #define TERRAIN_MESH_NAME "TerrainMesh"
    #define TERRAIN_ENT_NAME "TerrainEntity"
    #define TERRAIN_MAT_NAME "Examples/GrassFloor"

            if (!tMesh)
            {
                tMesh = new TerrainMesh (TERRAIN_MESH_NAME, TERRAIN_MAT_NAME);
                tMesh->init (extents, // num of x points
                            extents, // num of z points
                            1000.0f, // scalex
                            1000.0f); // scalez
            }
            mEnt = SceneMgr->createEntity(entname, TERRAIN_MESH_NAME);

            mEnt->setMaterialName (TERRAIN_MAT_NAME);
            mEnt->setCastShadows(false);

		    // Generate geometry
            mheightData = new float[(extents+1)*(extents+1)];
		    Update(entname, extents);

		    // Add to scene
            mSn = SceneMgr->getRootSceneNode()->createChildSceneNode(Vector3(-500,0,-500));
		    mSn->attachObject(mEnt);

	    }
        //----------------------------------------------------------------------
        void Update(const String &entname, int extents)
	    {
            const uint side = static_cast <uint> (extents + 1);
            uint pos_offset = 0;
            const float scale = yscale;
            const int genBasis = ibasis;

            float cursor_move = 1000.0f / extents;
            for (float z = -500.0f; z <= 500.0f;)
            { 
                for (float x = 500.0f; x >= -500.0f;)
                {
                    // update positions
			        mheightData[pos_offset++] = scale * GetY(x, z, genBasis);
                    x -= cursor_move;
                } 
                z += cursor_move;
		    } 
   
          

            tMesh->update(mheightData); 
	    }
        //----------------------------------------------------------------------
        void UpdateFault(const String &entname, int extents)
	    {
            const uint side = static_cast <uint> (extents + 1);  
            switch (ibasis)
            {   
                case 6: 
                    erfau.MakeTerrainPlasma (mheightData,
                                            side, 
                                            0.8f,//rough
                                            yscale);
                    break;
                case 7:
                    erfau.MakeTerrainDeposition (mheightData,
                                            side,
                                            10,//jumps,
                                            100,//peakWalk, 
                                            10,//minParticlesPerJump, 
                                            100,//maxParticlesPerJump,
                                            0.5f, //caldera
                                            yscale);
                    break;
                case 8:
                    erfau.MakeTerrainFault (mheightData,
                                    side, 
                                    100,//iteration
                                    100, //maxdelta
                                    0,//mindelta
                                    4,//iterations Per Erosion Filter  (3 means each 3 filter erosion occurs)
                                    0.2f,// erosion roughness
                                    yscale);
                    break;
                case 9:
                    //for (uint i = 0; i < 1; ++i)
                    { 
                        erfau.Erode(mheightData, 
                                    side,
                                    0.2f);// erosion filter
                    } 
                    break;
            
            
            } // switch (ibasis)
            tMesh->update(mheightData);
	    }
        //----------------------------------------------------------------------

        void UpdateWaterErode(const String &entname, int extents)
        {
            const uint side = static_cast <uint> (extents + 1);  
            if (ibasis == 10)
            {        
                for (int i=0; i<2; i++)
                  watero.Erode(mheightData, side, 14 /*16*/);// water erosion filter
            }
            tMesh->update(mheightData);
        }
        //----------------------------------------------------------------------

        void SaveMap(const String &entname, const String &Filename)
	    {		

            const AxisAlignedBox mmbox = mEnt->getMesh()->getBounds();
            const Real maxy = mmbox.getMaximum().y;
            const Real miny = mmbox.getMinimum().y;
            const Real divider = 1.0f / (maxy - miny) * 256;
            uint image_size = (mExtent+1)*(mExtent+1);
            uchar *data = new uchar[image_size];   
            for (uint i = 0; i < image_size; ++i )           
            {
                data[i] = static_cast <uchar>  ((mheightData[i] - miny) * divider);
            }
            Image grayheightmap;
            grayheightmap.loadDynamicImage (data,
                                            static_cast <size_t> (mExtent+1),
                                            static_cast <size_t> (mExtent+1),
                                            1,
                                            PF_L8,
                                            true);
            grayheightmap.save(Filename); 
	    } 
        //----------------------------------------------------------------------
        void SaveRawMap(const String &entname, const String &Filename)
	    {	
            const AxisAlignedBox mmbox = mEnt->getMesh()->getBounds();
            const Real maxy = mmbox.getMaximum().y;
            const Real miny = mmbox.getMinimum().y;
            const Real divider = 1.0f / (maxy - miny) * 65535;
            uint image_size = (mExtent+1)*(mExtent+1);
            uchar *data = new uchar[image_size*2];   
            uint j = 0;
            for (uint i = 0; i < image_size; ++i )           
            {
                ushort syn = ushort  ((mheightData[i]  - miny)* divider);
                #if OGRE_ENDIAN == OGRE_ENDIAN_BIG
                    data[j] = uchar ((syn >> 8) & 0xff);
		            data[j+ 1] = uchar (syn & 0xff);
                #else
		            data[j] = uchar (syn & 0xff);
		            data[j+ 1] = uchar ((syn >> 8) & 0xff);
                #endif
                j += 2;      
            } // for (uint i = 0; i < image_size; ++i )

            
            DataStreamPtr image_chunk(new MemoryDataStream ((void*)data,
                                                            image_size * 2 * sizeof (uchar), 
                                                            true)); 

            std::ofstream outfile;
            outfile.open (const_cast< char * > (Filename.c_str()),
                std::ios::binary);
            // Write out
            outfile << image_chunk->getAsString ();
            outfile.close ();
        }
        //----------------------------------------------------------------------
        inline float GetY(const float x, const float z, const int basgen)  const 
	    {
		    float farr[2];

		    farr[0] = (fzoom * x) + xoff;		// Get X
		    farr[1] = (fzoom * z) + zoff;		// Get Z

		    switch (basgen) 
            {
			    case 0:
				    return frac.fBm(farr, foctaves);
			    case 1:
				    return frac.Turbulence(farr, foctaves);
			    case 2:
				    return frac.Multifractal(farr, foctaves, foffset);
			    case 3:
				    return frac.Heterofractal(farr, foctaves, foffset);
			    case 4:
				    return frac.HybridMultifractal(farr, foctaves, foffset, fgain);
			    case 5:
				    return frac.RidgedMultifractal(farr, foctaves, foffset, fgain);
		    } // switch (ibasis)
            return 0.0f;
	    }

};
