#ifndef _Flag_Mesh_H
#define _Flag_Mesh_H

#include "Ogre.h"

namespace Ogre
{
    #define TERRAINMESH_VERTEX_BINDING 0
    #define TERRAINMESH_NORMAL_BINDING 1
    #define TERRAINMESH_TEXCOORDBINDING 2

  /**  
   * \author Tuan Kuranes
   *
   * \par license
   * This code is absolutely free to use and modify. The code is provided "as is" with
   * no expressed or implied warranty. The author accepts no liability if it causes
   * any damage to your computer, causes your pet to fall ill, increases baldness
   * or makes your car start emitting strange noises when you start it up.
   * This code has no bugs, just undocumented features!
   * 
   * \todo 
   *     LOD support
   *     Prevent self_Intersection  
   * \bug 
   *     
   */
    class TerrainMesh
    {

        public:

            TerrainMesh(const String& MeshName, const String& MaterialName);
            ~TerrainMesh();

            /**
             *
             * \param t 
             *         Time since last update.
             * \param &gravity 
             *         Set Gravity that affects flag
             * \param &wind 
             *         Set wind under which Flag is
             * \return 
             */
            bool update (const float * const heightData);
            /**
             *
             * \param numvertX 
             * number of vertices on the X axe
             * \param numvertZ 
             * number of vertices on the Z axe
             * \param TerrainSizeX 
             * x scale 
             * \param TerrainSizeZ 
             *  z scale
             * \return 
             * true if all ok
             */
             bool init(const uint numvertX,  const uint numvertZ, 
                           const Real TerrainSizeX, const Real TerrainSizeZ);
                
            void setMaterialName (const String &m);

             /**
            * Update normals for Light And Wind each Frame
            * (Real Normal Calculation)
            */
            void calculateNormals( const float * const heightmap);

        private:

            SubMesh     *mTerrainMesh;
            String      mMaterialName;

            MeshPtr     mMesh ;

            uint        mNumvertZ;
            uint        mNumvertX;

            Real        mScaleZ;
            Real        mScaleX;

            bool        mInit;

            size_t      mVec3Buffsize;

            uint         mVertexCount;
            uint         mNumFaces;

            VertexData  *mVertexData;
            Vector3     *mVertexBuffer;
            Vector3     *mNormalBuffer;
            Vector2     *mTexcoordsBuffer;

            String      meshName;


            void initBuffers();

    };
}
#endif //_Flag_Mesh_H