/***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU Lesser General Public License as        *
*   published by the Free Software Foundation; either version 2 of the    *
*   License, or (at your option) any later version.                       *
*                                                                         *
***************************************************************************/
/**
@file 
WaterErode.h
@brief
Code that process a water erosion iterative model on a heightmap
@author
sputnick
@date
25/mar/2005

*/
#ifndef WaterErode_H
#define WaterErode_H


#include "Ogre.h"
#include "OgreImage.h"
#include "OgreString.h"
#include "OgreSingleton.h"


namespace Ogre
{


class WaterErode
{
    public :

        WaterErode();
        ~WaterErode();

        /**
        * Make the heighfield uses the max of information range
        * does modify heavily heightmap
        * \param field the heightmap
        * \param size field must be square and side is length of a side
        * \param hmin min height in heightfield
        * \param hmax max height in heightfield
        * \param maxColor 256 for 8bits heightmap and 65535 for raw files
        */
        void equalize (Real * const field, 
                        const uint size, 
                        const Real hmin, 
                        const Real hmax, 
                        const uint maxColor) const;

        void normalize (Real * const field, 
                        const uint size,
                        const float scale) const;
       
            
        /*
            Water Erosion filter -

        */
        void Erode(Real* const afMap, const uint size,const uint iNbIteration) const;
        
 private:
   
        void getNeighborsDelta(Real* const afMap, const uint size, uint u32Offset, double *adNeighborsDelta) const;
};
}
#endif //WaterErode_H
