/***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU Lesser General Public License as        *
*   published by the Free Software Foundation; either version 2 of the    *
*   License, or (at your option) any later version.                       *
*                                                                         *
***************************************************************************/
/**
@file 
ErodeFault.h
@brief
Utility that split big map in tiles, pre-calculate splatting by
pre-calculating coverage map and pre-calculate normals.
*/
#ifndef ErodeFault_H
#define ErodeFault_H


#include "Ogre.h"
#include "OgreImage.h"
#include "OgreString.h"
#include "OgreSingleton.h"


namespace Ogre
{

/**
 * \ingroup Tool_ErodeFault
 * \author tuan.kuranes@free.fr
 */
class ErodeFault
{
    public :

        ErodeFault();
        ~ErodeFault();

        /**
        * Make the heighfield uses the max of information range
        * does modify heavily heightmap
        * \param field the heightmap
        * \param size field must be square and side is length of a side
        * \param hmin min height in heightfield
        * \param hmax max height in heightfield
        * \param maxColor 256 for 8bits heightmap and 65535 for raw files
        */
        void equalize (Real * const field, 
                        const uint size, 
                        const Real hmin, 
                        const Real hmax, 
                        const uint maxColor) const;

        void normalize (Real * const field, 
                        const uint size,
                        const float scale) const;
        /*
          Generate terrain using diamond-square (plasma) algorithm
        */
        void MakeTerrainPlasma(Real * const field, 
                                const uint size, 
                                const float rough,
                                const float scale) const;
        /*
         Drop particles, peak, flood fill, erode
        */
        void MakeTerrainDeposition(Real * const field,
                                    const uint size,
                                    const uint jumps,
                                    const uint peakWalk, 
                                    const uint minParticlesPerJump,
                                    const uint maxParticlesPerJump,
                                    const float caldera,
                                    const float scale) const;
        
        void MakeTerrainFault(Real * const field, 
                                const uint size,
                                const uint iterations,
                                const uint maxDelta, 
                                const uint minDelta,
                                const uint iterationsPerFilter,
                                const float filter,
                                const float scale) const;      
        /*
            Erosion filter -
            Erodes a terrain in all 4 directions
        */
        void Erode(Real * const field, const uint size, const float filter) const;
        
 private:
   
        /*
        Returns a random number between v1 and v2
        */
        float RangedRandom(const float v1, const float v2) const;         
        /*
        Erosion filter -
            FilterHeightBand applies a FIR filter across a row or column of the height field
        */ 
        void ErodeOneDirection(float * const band, 
            const int stride,  
            const uint count, 
            const float filter) const;
};
}
#endif //ErodeFault_H
